#!/home/hamaji/local/bin/gosh

(use text.html-lite)
(use www.cgi)
(use math.mt-random)
(use gauche.collection)
(use srfi-1)

(define mt (make <mersenne-twister> :seed (sys-time)))

(define ids (iota 16))

(define (db-file num)
  (string-append "db/" (number->string num) ".db"))

(define (load-game num)
  (eval (read (open-input-file (db-file num))) (null-environment 5)))

(define (save-game game num)
  (let ((oport (open-output-file (db-file num))))
    (write-char #\' oport)
    (write game oport)
    (close-output-port oport)
    ))

(define (pnt->string pnt)
  (string-join (map number->string pnt) ","))

(define (update-board board id x y)
  (map (lambda (i)
         (if (= i id)
             (list x y)
             (list-ref board i))) ids)
  )

(define (board-x board id)
  (list-ref (list-ref board id) 0))

(define (board-y board id)
  (list-ref (list-ref board id) 1))

(define (find-board board x y)
  (find
   (lambda (i)
     (and (= (board-x board i) x)
          (= (board-y board i) y)))
   ids)
  )

(define (died-num board p)
  (let ((y (if (= p 0) -1 6)))
    (length (filter (lambda (i) (= (board-y board i) y)) ids)))
  )

(define (died-good-num board p)
  (let ((y (if (= p 0) -1 6)))
    (length (filter (lambda (i) (= (board-y board i) y))
                    (iota 4 (if (= p 0) 8 0)))))
  )

(define (died-evil-num board p)
  (let ((y (if (= p 0) -1 6)))
    (length (filter (lambda (i) (= (board-y board i) y))
                    (iota 4 (if (= p 0) 12 4)))))
  )

(define (main args)
  (cgi-main
   (lambda (params)
     (let ((cmd (cgi-get-parameter "cmd" params)))
       (if cmd
           (let ()
             (if (string=? cmd "init")
                 (let ()
                   (define (init l)
                     (let ((x (+ (mt-random-integer mt 4) 1))
                           (y (+ (mt-random-integer mt 2)
                                 (if (< (length l) 8) 0 4))))
                       (if (find (lambda (p)
                                   (and (= (car p) x)
                                        (= (car (cdr p)) y))) l)
                           (init l)
                           (let ((nextl (append! l (list (list x y)))))
                             (if (= (length nextl) 16)
                                 nextl
                                 (init nextl))))))
                   (save-game (append (list 0) (init (list))) 100)
                   ))
             (if (string=? cmd "setup")
                 (let* ((game (load-game 100))
                        (state (car game))
                        (board (cdr game))
                        (p (string->number (cgi-get-parameter "p" params)))
                        (nexts (+ state p))
                        )
                   (save-game
                    (append (list (if (< nexts 4) nexts state)) board) 100)
                   ))
             (if (string=? cmd "move")
                 (let* ((game (load-game 100))
                        (state (car game))
                        (board (cdr game))
                        (x (string->number (cgi-get-parameter "x" params)))
                        (y (string->number (cgi-get-parameter "y" params)))
                        (id (string->number(cgi-get-parameter "id" params)))
                        (target (find-board board x y))
                        )
                   (if target
                       (let ((p (quotient id 8)))
                         (set! board
                               (update-board board target
                                             (died-num board p)
                                             (if (= p 0) -1 6)))
                         (if (= 4 (died-good-num board p))
                             (set! state (+ 5 p)))
                         (if (= 4 (died-evil-num board p))
                             (set! state (- 6 p)))
                         ))
                   (if (and
                        (= 3 state)
                        (let ((g1 (find-board board 0 0))
                              (g2 (find-board board 5 0)))
                          (or
                           (and g1 (and (>= g1 8) (< g1 12)))
                           (and g2 (and (>= g2 8) (< g2 12))))
                          ))
                       (set! state 6))
                   (if (and
                        (= 4 state)
                        (let ((g1 (find-board board 0 5))
                              (g2 (find-board board 5 5)))
                          (or
                           (and g1 (and (>= g1 0) (< g1 4)))
                           (and g2 (and (>= g2 0) (< g2 4))))
                          ))
                       (set! state 5))
                   (save-game
                    (append (list (if (= state 3) 4 (if (= state 4) 3 state)))
                            (update-board board id x y)) 100)
                   ))
             )
           )
       `(,"Content-Type: text/plain\r\nCache-Control: no-cache\r\n\r\n"
         ,(let* ((game (load-game 100))
                 (state (car game))
                 (board (cdr game)))
            (string-append
             "state=" (number->string state) "&"
             (string-join
              (map (lambda (i x)
                     (string-append
                      "g" (number->string i) "=" (pnt->string x)))
                   ids board) "&")))
         )))))
